defmodule Clickhouse do
  @moduledoc """
  A Clickhouse client.

  This currently represents an early work in progress.
  """

  @type name :: atom()
  @type option ::
          {:name, name()}
          | {:interface, Clickhouse.Interface.t()}
          | {:interface_opts, keyword()}
  @type options :: [option()]

  use Supervisor

  alias Clickhouse.{Error, Format, Interface, Query}

  @opts_schema %{
    name: [type: :atom, required: true, default: :default],
    interface: [type: :module, required: true, default: Clickhouse.Interface.HTTP],
    interface_opts: [type: :list, required: true, default: []]
  }

  ################################
  # Public API
  ################################

  @doc """
  Starts a Clickhouse client.

  ## Options

    * `:name` - The name of the client - defaults to `:default`.
    * `:interface` - The interface to use for queries - defaults to `Clickhouse.Interface.HTTP`.
    * `:interface_opts` - Additional keyword options to pass to the interface.
  """
  @spec start_link(options()) :: Supervisor.on_start()
  def start_link(opts \\ []) do
    with {:ok, opts} <- validate_opts(opts, @opts_schema) do
      Supervisor.start_link(__MODULE__, opts)
    end
  end

  @doc """
  Executes a query using a Clickhouse client.
  """
  @spec query(name(), Clickhouse.Query.st()) ::
          {:ok, Clickhouse.Result.t()} | {:error, Clichouse.Error.t()}
  def query(name, statement, params \\ []) do
    with {:ok, query} <- prepare(name, statement, params),
         {:ok, result} <- execute(query) do
      decode(result)
    end
  end

  ################################
  # Supervisor Callbacks
  ################################

  @impl Supervisor
  def init(opts) do
    init_config(opts)

    children = [
      build_interface(opts)
    ]

    Supervisor.init(children, strategy: :one_for_one)
  end

  ################################
  # Private API
  ################################

  defp validate_opts(opts, schema) do
    case KeywordValidator.validate(opts, schema) do
      {:ok, _} = result -> result
      {:error, reason} -> {:stop, reason}
    end
  end

  defp init_config(opts) do
    name = Keyword.fetch!(opts, :name)
    table_name = table__name(name)
    interface = Keyword.fetch!(opts, :interface)

    :ets.new(table_name, [:named_table, read_concurrency: true])
    :ets.insert(table_name, {:interface, interface})
  end

  defp config(name, key) do
    [{^key, value}] = :ets.lookup(table__name(name), key)
    {:ok, value}
  rescue
    ArgumentError ->
      {:error, Error.new(:not_found, "Clickhouse #{inspect(name)} not found.")}
  end

  defp table__name(name) do
    :"#{__MODULE__}.#{name}"
  end

  defp build_interface(opts) do
    name = Keyword.fetch!(opts, :name)
    interface = Keyword.fetch!(opts, :interface)
    interface_opts = Keyword.fetch!(opts, :interface_opts)
    interface_opts = Keyword.put(interface_opts, :name, name)

    {interface, interface_opts}
  end

  defp prepare(name, statement, params) do
    with {:ok, interface} <- config(name, :interface) do
      query = Query.new(name, interface, statement, params)
      Format.encode(query)
    end
  end

  defp execute(query), do: Interface.execute(query)

  defp decode(result), do: Format.decode(result)
end
