if Code.ensure_loaded?(NimbleCSV) do
  defmodule Clickhouse.Format.TSV do
    @moduledoc """
    An implementation of the Clickhouse `"TSV"`/`"TabSeparatedValues"` format.
    """

    @behaviour Clickhouse.Format

    alias Clickhouse.{Error, Query, Result}

    @parser __MODULE__.Parser

    NimbleCSV.define(@parser, separator: "\t", escape: "\"")

    ################################
    # Clickhouse.Format Callbacks
    ################################

    @impl Clickhouse.Format
    @spec decode(Clickhouse.Result.t()) :: {:ok, Clickhouse.Result.t()}
    def decode(result) do
      data = @parser.parse_string(result.__raw__, skip_headers: false)
      result = Result.put_data(result, data)
      {:ok, result}
    end

    @impl Clickhouse.Format
    @spec encode(Clickhouse.Query.t()) ::
            {:ok, Clickhouse.Query.t()} | {:error, Clickhouse.Error.t()}
    def encode(%Query{type: :insert, params_count: count}) when count > 1 do
      error =
        Error.new(
          :encode_error,
          "TSV insert statement can only support a single '?'."
        )

      {:error, error}
    end

    def encode(%Query{type: :insert, params_count: 0, params: [_ | _]}) do
      error =
        Error.new(
          :encode_error,
          "TSV insert statement with params must include a single '?'."
        )

      {:error, error}
    end

    def encode(%Query{type: :insert, params_count: 1, params: params} = query) do
      data = do_encode(params)
      encoded = String.replace(query.statement, "?", data)
      {:ok, Query.put_statement(query, encoded)}
    end

    def encode(query) do
      {:ok, query}
    end

    ################################
    # Private API
    ################################

    defp do_encode(params) do
      params
      |> Enum.map(&encode_row/1)
      |> @parser.dump_to_iodata()
      |> IO.iodata_to_binary()
    end

    defp encode_row(values), do: Enum.map(values, &encode_value/1)

    defp encode_value(nil), do: nil
    defp encode_value(false), do: 0
    defp encode_value(true), do: 1
    defp encode_value(string) when is_binary(string), do: string
    defp encode_value(integer) when is_integer(integer), do: to_string(integer)
    defp encode_value(atom) when is_atom(atom), do: to_string(atom)
    defp encode_value(%Date{} = date), do: to_string(date)

    defp encode_value(list) when is_list(list) do
      values = Enum.map(list, &encode_array_value/1)
      "[" <> Enum.join(values, ",") <> "]"
    end

    defp encode_array_value(nil), do: nil
    defp encode_array_value(false), do: 0
    defp encode_array_value(true), do: 1
    defp encode_array_value(string) when is_binary(string), do: "'" <> string <> "'"
    defp encode_array_value(integer) when is_integer(integer), do: to_string(integer)
    defp encode_array_value(atom) when is_atom(atom), do: to_string(atom)
    defp encode_array_value(%Date{} = date), do: to_string(date)
  end
end
