defmodule Clickhouse.Interface.HTTP do
  @moduledoc """
  An interface to interact with a Clickhouse server via HTTP.
  """

  @behaviour Clickhouse.Interface

  use GenServer

  alias Clickhouse.{Error, Format, Query, Result}

  @opts_schema %{
    name: [type: :atom, required: true, default: :default],
    urls: [type: {:list, :binary}, required: true, default: ["http://localhost:8123"]]
  }

  ################################
  # Clickhouse.Interface Callbacks
  ################################

  @doc """
  Starts the HTTP interface server.
  """
  @impl Clickhouse.Interface
  @spec start_link(keyword()) :: GenServer.on_start()
  def start_link(opts) do
    with {:ok, opts} <- validate_opts(opts, @opts_schema) do
      GenServer.start_link(__MODULE__, opts)
    end
  end

  @doc """
  Executes a query via HTTP.
  """
  @impl Clickhouse.Interface
  @spec execute(Clickhouse.Query.t()) ::
          {:error, Clickhouse.Error.t()} | {:ok, Clickhouse.Result.t()}
  def execute(query) do
    with {:ok, urls} <- config(query.__name__, :urls) do
      statement = Query.get_statement(query)
      do_execute(urls, statement)
    end
  end

  ################################
  # GenServer Callbacks
  ################################

  @impl GenServer
  def init(opts) do
    with {:ok, opts} <- validate_opts(opts, @opts_schema) do
      state = init_state(opts)
      init_config(state)
      {:ok, state}
    end
  end

  ################################
  # Private API
  ################################

  defp validate_opts(opts, schema) do
    case KeywordValidator.validate(opts, schema) do
      {:ok, _} = result -> result
      {:error, reason} -> {:stop, reason}
    end
  end

  defp init_state(opts) do
    name = Keyword.get(opts, :name)

    %{
      name: name,
      table_name: table_name(name),
      urls: Keyword.get(opts, :urls)
    }
  end

  defp init_config(state) do
    :ets.new(state.table_name, [:named_table, read_concurrency: true])
    :ets.insert(state.table_name, {:urls, state.urls})
  end

  defp table_name(name) do
    :"#{__MODULE__}.#{name}"
  end

  defp config(name, key) do
    [{^key, value}] = :ets.lookup(table_name(name), key)
    {:ok, value}
  rescue
    ArgumentError ->
      {:error, Error.new(:not_found, "HTTP interface #{inspect(name)} not found.")}
  end

  defp do_execute([url], statement) do
    do_execute(url, statement)
  end

  defp do_execute(urls, statement) when is_list(urls) do
    url = Enum.random(urls)
    do_execute(url, statement)
  end

  defp do_execute(url, statement) when is_binary(url) do
    case Mojito.post(url, [], statement) do
      {:ok, %{status_code: 200} = response} ->
        build_result(response)

      {:ok, %{status_code: 400} = response} ->
        {:error, Error.new(:bad_request, response.body)}

      {:ok, %{status_code: 404} = response} ->
        {:error, Error.new(:not_found, response.body)}

      {:ok, %{status_code: code} = response} when code >= 500 ->
        {:error, Error.new(:server_error, response.body)}

      {:error, %{reason: %{reason: reason}}} ->
        {:error, Error.new(reason, nil)}

      {:error, %{reason: reason, message: message}} ->
        {:error, Error.new(reason, message)}
    end
  end

  defp build_result(response) do
    with {:ok, format} <- parse_format(response.headers) do
      result = Result.new(response.body, format)
      {:ok, result}
    end
  end

  defp parse_format([]), do: {:ok, nil}

  defp parse_format([{"x-clickhouse-format", format} | _headers]) do
    Format.find(format)
  end

  defp parse_format([_ | headers]) do
    parse_format(headers)
  end
end
