defmodule Clickhouse.Query do
  @moduledoc """
  A Clickhouse query.
  """

  alias Clickhouse.Format

  @type statement :: binary()
  @type params :: list()
  @type t :: %__MODULE__{}

  @format_regex ~r/(?<= (?:FORMAT|format) )(\w+)/
  @create_regex ~r/\bCREATE\b/i
  @select_regex ~r/\bSELECT\b/i
  @insert_regex ~r/\bINSERT\b/i
  @alter_regex ~r/\bALTER\b/i

  @derive {Inspect, only: [:type, :statement, :params, :params_count]}
  defstruct [
    :type,
    :statement,
    :params,
    :params_count,
    :__name__,
    :__format__,
    :__interface__,
    :__statement__
  ]

  ################################
  # Public API
  ################################

  @spec new(Clickhouse.name(), Clickhouse.Interface.t(), statement(), params()) ::
          Clickhouse.Query.t()
  def new(name, interface, statement, params \\ []) do
    %__MODULE__{
      __name__: name,
      __interface__: interface,
      __format__: parse_format(statement),
      type: parse_type(statement),
      statement: statement,
      params: params,
      params_count: parse_params_count(statement)
    }
  end

  @spec put_statement(Clickhouse.Query.t(), binary()) :: Clickhouse.Query.t()
  def put_statement(query, statement) do
    %{query | __statement__: statement}
  end

  @spec get_statement(Clickhouse.Query.t()) :: binary()
  def get_statement(query) do
    query.__statement__ || query.statement
  end

  ################################
  # Private API
  ################################

  defp parse_format(statement) do
    case Regex.run(@format_regex, statement, capture: :first) do
      [format] -> find_format(format)
      _ -> nil
    end
  end

  defp find_format(format) do
    case Format.find(format) do
      {:ok, format} -> format
      _ -> nil
    end
  end

  defp parse_type(statement) do
    with {_, false} <- {:create, Regex.match?(@create_regex, statement)},
         {_, false} <- {:select, Regex.match?(@select_regex, statement)},
         {_, false} <- {:insert, Regex.match?(@insert_regex, statement)},
         {_, false} <- {:alter, Regex.match?(@alter_regex, statement)} do
      :unknown
    else
      {type, true} -> type
    end
  end

  defp parse_params_count(statement) do
    statement
    |> String.codepoints()
    |> Enum.count(fn s -> s == "?" end)
  end
end
